'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

const Promise = require('bluebird');
const Errors = require('common-errors');
const setMetadata = require('../utils/updateMetadata.js');
const scrypt = require('../utils/scrypt.js');
const redisKey = require('../utils/key.js');
const emailValidation = require('../utils/send-email.js');
const jwt = require('../utils/jwt.js');
const uuid = require('node-uuid');

var _require = require('../constants.js');

const USERS_INDEX = _require.USERS_INDEX;
const USERS_DATA = _require.USERS_DATA;
const USERS_ACTIVE_FLAG = _require.USERS_ACTIVE_FLAG;
const MAIL_REGISTER = _require.MAIL_REGISTER;

const isDisposable = require('../utils/isDisposable.js');
const mxExists = require('../utils/mxExists.js');
const makeCaptchaCheck = require('../utils/checkCaptcha.js');
const userExists = require('../utils/userExists.js');
const aliasExists = require('../utils/aliasExists.js');
const noop = require('lodash/noop');
const assignAlias = require('./alias.js');

/**
 * Verify ip limits
 * @param  {redisCluster} redis
 * @param  {Object} registrationLimits
 * @param  {String} ipaddress
 * @return {Function}
 */
function checkLimits(redis, registrationLimits, ipaddress) {
  var _registrationLimits$i = registrationLimits.ip;
  const time = _registrationLimits$i.time;
  const times = _registrationLimits$i.times;

  const ipaddressLimitKey = redisKey('reg-limit', ipaddress);
  const now = Date.now();
  const old = now - time;

  return function iplimits() {
    return redis.pipeline().zadd(ipaddressLimitKey, now, uuid.v4()).pexpire(ipaddressLimitKey, time).zremrangebyscore(ipaddressLimitKey, '-inf', old).zcard(ipaddressLimitKey).exec().then(props => {
      const cardinality = props[3][1];
      if (cardinality > times) {
        const msg = 'You can\'t register more users from your ipaddress now';
        throw new Errors.HttpStatusError(429, msg);
      }
    });
  };
}

/**
 * Creates user with a given hash
 */
function createUser(redis, username, activate, deleteInactiveAccounts, userDataKey) {
  /**
   * Input from scrypt.hash
   */
  return function create(hash) {
    const pipeline = redis.pipeline();

    pipeline.hsetnx(userDataKey, 'password', hash);
    pipeline.hsetnx(userDataKey, USERS_ACTIVE_FLAG, activate);

    return pipeline.exec().spread(function insertedUserData(passwordSetResponse) {
      if (passwordSetResponse[1] === 0) {
        throw new Errors.HttpStatusError(412, `User "${ username }" already exists`);
      }

      if (!activate && deleteInactiveAccounts >= 0) {
        // WARNING: IF USER IS NOT VERIFIED WITHIN <deleteInactiveAccounts>
        // [by default 30] DAYS - IT WILL BE REMOVED FROM DATABASE
        return redis.expire(userDataKey, deleteInactiveAccounts);
      }

      return null;
    });
  };
}

/**
 * Registration handler
 * @param  {Object} message
 * @return {Promise}
 */
module.exports = function registerUser(message) {
  const redis = this.redis;
  const config = this.config;
  const deleteInactiveAccounts = config.deleteInactiveAccounts;
  const captchaConfig = config.captcha;
  const registrationLimits = config.registrationLimits;

  // message

  const username = message.username;
  const alias = message.alias;
  const password = message.password;
  const audience = message.audience;
  const ipaddress = message.ipaddress;
  const skipChallenge = message.skipChallenge;
  const activate = message.activate;

  const captcha = message.hasOwnProperty('captcha') ? message.captcha : false;
  const metadata = message.hasOwnProperty('metadata') ? message.metadata : false;

  // task holder
  const logger = this.log.child({ username: username, action: 'register' });

  // make sure that if alias is truthy then activate is also truthy
  if (alias && !activate) {
    throw new Errors.HttpStatusError(400, 'Account must be activated when setting alias during registration');
  }

  let promise = Promise.bind(this, username);

  // optional captcha verification
  if (captcha) {
    logger.debug('verifying captcha');
    promise = promise.tap(makeCaptchaCheck(redis, username, captcha, captchaConfig));
  }

  if (registrationLimits) {
    if (registrationLimits.noDisposable) {
      promise = promise.tap(isDisposable(username));
    }

    if (registrationLimits.checkMX) {
      promise = promise.tap(mxExists(username));
    }

    if (registrationLimits.ip && ipaddress) {
      promise = promise.tap(checkLimits(redis, registrationLimits, ipaddress));
    }
  }

  // shared user key
  const userDataKey = redisKey(username, USERS_DATA);

  // step 2, verify that user _still_ does not exist
  promise = promise
  // verify user does not exist at this point
  .tap(userExists).throw(new Errors.HttpStatusError(409, `"${ username }" already exists`)).catchReturn({ statusCode: 404 }, username).tap(alias ? aliasExists(alias, true) : noop)
  // step 3 - encrypt password
  .then(() => {
    if (password) {
      return password;
    }

    // if no password was supplied - we auto-generate it and send it to an email that was provided
    // then we hash it and store in the db
    return emailValidation.send.call(this, username, MAIL_REGISTER).then(ctx => ctx.context.password);
  }).then(scrypt.hash)
  // step 4 - create user if it wasn't created by some1 else trying to use race-conditions
  .then(createUser(redis, username, activate, deleteInactiveAccounts, userDataKey))
  // step 5 - save metadata if present
  .return({
    username: username,
    audience: audience,
    metadata: {
      $set: _extends({
        username: username
      }, metadata || {})
    }
  }).then(setMetadata).return(username);

  // no instant activation -> send email or skip it based on the settings
  if (!activate) {
    return promise.then(skipChallenge ? noop : emailValidation.send).return({ requiresActivation: true });
  }

  // perform instant activation
  return promise
  // add to redis index
  .then(() => redis.sadd(USERS_INDEX, username))
  // call hook
  .return(['users:activate', username, audience]).spread(this.hook)
  // assign alias if specified
  .tap(() => {
    if (!alias) {
      return null;
    }

    // adds on-registration alias to the user
    return assignAlias.call(this, { username: username, alias: alias });
  })
  // login user
  .return([username, audience]).spread(jwt.login);
};